 /*
    CLM_uah - Constrained Local Model (CLM) sample code. A (partial)
    implementation of D. Christinacce and T. Cootes algorithm.

    Jesus Nuevo-Chiquero, University of Alcal. 2010
    jesus.nuevochiquero@gmail.com

    This file is part of 'CLM_uah'.

    CLM_uah is free software: you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    CLM_uah is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with CLM_uah.  If not, see <http://www.gnu.org/licenses/>.

*/
#include <iostream>
#include <exception>


#include "Shape.hh"


using namespace std;




int ShapeBase::load(const string file)
{

  FileStorage fs(file, FileStorage::READ);

    if(fs.isOpened()){
	FileNode node = fs["pdm_model"];
	if(!node.empty()){
	    try{
		read(node["pdm_mean"], mean);
		read(node["pdm_vectors"], eigenvectors);
		read(node["pdm_eigvals"], eigenvalues);
	    }
	    catch(Exception e){
		cerr << e.err <<endl;
		return -1;
	    }

	    sqrt(eigenvalues.t(), m_eig_sqrt);
	}
    }
    else
	return -1;

    return 0;

}

void Shape::init(void)
{

    m_bs = Mat::zeros(1, m_base.eigenvectors.rows, m_base.eigenvectors.type());
    m_st = Mat::zeros(1,4,m_bs.type());
    

}

void Shape::set_parameters(const Mat bs, const Mat st)
{


    assert(bs.cols == m_base.eigenvectors.rows);
    assert(st.cols == 4);

    if(!bs.empty()){
	if(countNonZero(bs == m_bs) == bs.cols && countNonZero(st == m_st) == m_st.cols){
	    return;
	}
    }

    bs.copyTo(m_bs);
    st.copyTo(m_st);
    m_updated = false;

}

void Shape::get_parameters(Mat& bs, Mat& st) const 
{

    m_bs.copyTo(bs);
    m_st.copyTo(st);

}

void Shape::set(const Mat pts, bool limit)
{

    Mat shape;

    if(pts.empty())
	throw ;

    pts.copyTo(shape);

    m_st = project_over_ST(shape, m_base);
    remove_ST(shape, m_st);

    //project over the PCA in ShapeBase

    Mat m = shape.reshape(0,1);

    m_bs = m_base.project(m);

    //limit the values
    if(limit){
	m_bs = min(m_bs, 3*m_base.m_eig_sqrt);
	m_bs = max(m_bs, -3*m_base.m_eig_sqrt);
    }
    m_updated = false;
}

Mat Shape::get() 
{

    Mat m;

    if(m_updated){
	m_pts.copyTo(m);
    }
    else{
	m = m_base.backProject(m_bs);
	
	m = m.reshape(0,2);
	
	double p0 = 1+ m_st.at<float>(0,0)*cos(m_st.at<float>(0,1));
	double p1 = m_st.at<float>(0,0)*sin(m_st.at<float>(0,1));

	Mat_<float> rot (2,2);
	rot.diag() = p0;
	rot.at<float>(0,1)= -p1;
	rot.at<float>(1,0) = p1;
	
	m = rot*m;

	Mat r = m.row(0);
	r += m_st.at<float>(0,2);
	r = m.row(1);
	r += m_st.at<float>(0,3);

	m.copyTo(m_pts);
	m_updated = true;
    }

    return m;
}

Mat project_over_ST(const Mat& sh, const ShapeBase& pdm)
{

    Mat s0 = pdm.mean.reshape(0,2);

    Mat proj = Mat::zeros(1,4, sh.type());

    //get the centre of the shape
      
    Mat r;

    reduce(sh, r, 1, CV_REDUCE_AVG);

    proj.at<float>(0,2) = r.at<float>(0,0);
    proj.at<float>(0,3) = r.at<float>(1,0);

    double p0 = sh.dot(s0) - 1;
    
    Mat s_ (s0.size(), s0.type());
    r = s_.row(1);
    s0.row(0).copyTo(r);

    r = s_.row(0);
    s0.row(1).copyTo(r);
    r = -1 *r;

    double p1 = sh.dot(s_);

    
    proj.at<float>(0,1) = atan2(p1, p0);
    
    proj.at<float>(0,0) = sqrt(p0*p0 + p1*p1);

    return proj;
}


void remove_ST(Mat sh, const Mat st_p)
{

    double p0, p1;

    p0 = st_p.at<float>(0,0) * cos(st_p.at<float>(0,1));
    p1 = st_p.at<float>(0,0) * sin(st_p.at<float>(0,1));

    sh.row(0) -= Scalar(st_p.at<float>(0,3));
    sh.row(1) -= Scalar(st_p.at<float>(0,4));


    Mat M (2,2,sh.type());
    M.diag() = 1 + p0;
    M.at<float>(0,1) = -1* p1;
    M.at<float>(1,0) = p1;

    sh = M.inv() * sh;

}
