//Compute the coupling matrix between the texture and the shape
//parameters (see CLM's paper for more details)

 /*
    CLM_uah - Constrained Local Model (CLM) sample code. A (partial)
    implementation of D. Christinacce and T. Cootes algorithm.

    Jesus Nuevo-Chiquero, University of Alcal. 2010
    jesus.nuevochiquero@gmail.com

    This file is part of 'CLM_uah'.

    CLM_uah is free software: you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    CLM_uah is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with CLM_uah.  If not, see <http://www.gnu.org/licenses/>.

*/
#include <stdio.h>
#include <time.h>

#include <sys/stat.h>
#include <sys/types.h>
#include <errno.h>
#include <stddef.h>
#include <dirent.h>

#include <vector>
#include <stdlib.h>
#include <algorithm>
#include <iostream>
#include <fstream>
#include <utility>


#include <stdexcept>

#include <getopt.h>

#include <cxcore.h>
#include <cv.h>
#include <highgui.h>

#include <boost/program_options.hpp>

#include "basic_utils.hh"
#include "bioid_parser.hh"
#include "Shape.hh"
#include "AppearanceBase.hh"

#include "config.h"

using namespace cv;
namespace po = boost::program_options;
using namespace std;


#define DEFAULT_KEEP_PCA 0.95
#define DEFAULT_MIN_STD_VALUE 1e-2

// const int DEFAULT_PATCH_WIDTH = 13;
// const int DEFAULT_PATCH_HEIGHT = DEFAULT_PATCH_WIDTH;

const char *const svn_str = "$Id: compute_coupling_matrix.cpp 799 2010-02-01 16:29:23Z jnuevo $";
const char *const svn_rev = "$Rev: 945 $";


typedef struct cliparams_t{
    bool use_bioid;
    
    double alpha;
    double keep_pca;    
    int max_components;

    string shape_file, app_file;   

    std::vector<std::string> inputs; //folders or list of images & marks
    string ofile;
}cliparams_t;



int parse_cli(int argc, char *argv[], cliparams_t& clarg)
{

  try{

	po::options_description generic("Generic options");
	generic.add_options()
	    ("version,v", "print version string")
	    ("help,h", "produce help message")    
	    ;

	po::options_description config("Configuration (Required)");
        config.add_options()
	    ("app_model,a",po::value<string>(&clarg.app_file), "Appearance model description file")
	    ("shape_model,s", po::value<string>(&clarg.shape_file), "Shape model (PDM) description file")
	    ;

	po::options_description output("Output");
	output.add_options()
	    ("variation,u",po::value<double>(&clarg.alpha)->default_value(0.3), "variantion on the shape parameters")
	    ("keep,k", po::value<double>(&clarg.keep_pca),
	     "Representation percentage to be kept when performing PCA. Default 0.95")
	    ("max_components,m", po::value<int>(&clarg.max_components)->default_value(0),
	     "Maximum number of vectors, regardless if they cover the representation percentage (0 = no limit)")
	    ("bioid,b",
	     "The image marks are given in the BioID point format used by D. Cristinacce (default is CvMat XML format)")
	    ("output,o", po::value<std::string>(&clarg.ofile), "Output file, with the weight matrix and the coupled PCA")
	    ;
	
	// Hidden options, will be allowed both on command line and
        // in config file, but will not be shown to the user.
        po::options_description hidden("Hidden options");
        hidden.add_options()
            ("input-files", po::value< std::vector<std::string> >(&clarg.inputs), "input files")
            ;
	
  	po::positional_options_description p;
        p.add("input-files", -1);

	po::options_description visible;
        visible.add(generic).add(config).add(output);

	po::options_description cmdline_options;
        cmdline_options.add(generic).add(config).add(hidden).add(output);


	po::variables_map vm;
        store(po::command_line_parser(argc, argv).
              options(cmdline_options).positional(p).run(), vm);
	
        notify(vm);
	
	if(vm.count("version")){	    
	    std::cout << "Version: "<<CLM_uah_VERSION_MAJOR<<"."
		      <<CLM_uah_VERSION_MINOR<<"."
		      <<CLM_uah_VERSION_PATCH<<std::endl;
	    std::cout << "SVN " <<svn_rev << std::endl;
	    return 1;
	}

	if (vm.count("help")) {
	    cout << "Compute the shape parameter weight matrix and the shape-texture parameters coupled PCA"<<endl;
	    std::cout << "Usage: computer_coupling_matrix options <image_files marks_files | image_folder marks_folder>"<<std::endl;
	    std::cout << visible << std::endl;
	    std::cout << "Note: if two folders are given as arguments,\n" 
		      << "\timages and marks are read sequentially\n"
		      << "\tfrom both folders. Make sure they correspond."<< std::endl
		      <<endl;
   
            return 1;
        }

	if(!vm.count("app_model")){
	    cerr << "ERROR: An appearance model file is needed"<<endl;
	    return -1;
	}

	if(!vm.count("shape_model")){
	    cerr << "ERROR: A shape model file is needed"<<endl;
	    return -1;
	}

	if(!vm.count("output")){
	    for(uint ai=0; ai<1e5;ai++){
		char file[128];
		sprintf(file, "coupled_base_%05d.xml", ai);
		
		if(access (file, F_OK)){
		    clarg.ofile = file;
		    break;
		}
	    }
	    if(!clarg.ofile.length()){
		std::cerr << "Output file could not be created, too many files in current folder"<<std::endl;
		exit(1);
	    }		
	}

	if(vm.count("bioid"))
	    clarg.use_bioid = true;
	else
	    clarg.use_bioid = false;
	if(!vm.count("keep")){
	    clarg.keep_pca = DEFAULT_KEEP_PCA;
	}

	if(!vm.count("variation")){
	    clarg.alpha = 0.1;
	}

	if(clarg.alpha<-1)
	    clarg.alpha = -1;
	else if(clarg.alpha>1)
	    clarg.alpha = 1;

  }
  catch(std::exception& e)
      {
	  std::cout << e.what() << "\n";
	  return 1;
      }    
    return 0;
}




int extract_patches(const Mat& img, Mat& points, std::vector<Mat> &patches,
		    const Size sz)
{


    unsigned int out = 0;

   

// #ifdef DEBUG
//     static int wait_time = 0;
//     Mat dbgimg (img.size(), CV_32F, 3);
//     cvtColor(img, dbgimg, CV_GRAY2BGR);
//     namedWindow("Extract_DBG", CV_WINDOW_AUTOSIZE);
// #endif    


    //extract the patches. The patches are centered around the marks
    
    Rect iRect (Point(0,0), img.size());//size of the image
    for(int ai=0; ai<points.cols;ai++){
	Mat p;

	//check if the patch is completely inside the image.
	Rect r (points.at<float>(0,ai) - sz.width/2,
		points.at<float>(1,ai) - sz.height/2,
		sz.width, sz.height);

	Rect maxR = iRect | r;

	if(maxR == iRect)//patch is in the image
	    p = img(r).clone();
	else
	    out++;

	patches.push_back(p);//pushed even if it is empty.

// #ifdef DEBUG
// 	rectangle(dbgimg, r.tl(), r.br(), CV_RGB(0,255,0));
// #endif
    }
    
     
// #ifdef DEBUG
 
//     imshow("Extract_DBG", dbgimg);
//     unsigned char key = static_cast<unsigned char> (waitKey(wait_time));
//     if(key == '0' || key == ' '){
// 	wait_time = 0;
//     }		
//     else if(key == '1'){
// 	wait_time = 2;
//     }
//     else if(key == '2'){
// 	wait_time = 10;
//     }
//     else if(key == '3'){
// 	wait_time = 100;
//     }
// #endif

    return patches.size() - out;
}



int savePCA(FileStorage &fs, const PCA& myPCA, const double keep_pca, int max = 0)
{

    int keepN = 0;
    
    Scalar total = sum(myPCA.eigenvalues);
    const Mat& diag = myPCA.eigenvalues;
    
    double pc = 0;

    if(max<=0)
	max = diag.rows;

    for( ; keepN<diag.rows && keepN<max; keepN++){
	pc += diag.at<float>(keepN,0);
	if(pc > keep_pca * total[0]){
	    // 			myPCA.eigenvectors(Range(0,keepN+1), Range::all()).copyTo(eigVectors);
// 			diag(Range(0,keepN+1), Range::all()).copyTo(eigValues);
	    break;
	}
    }

    if(keepN == max && pc < keep_pca*total[0]){
	std::cout << "Maximum number of components reached ("<<max<<"). Representation kept: "
		  << pc/total[0]<<std::endl;
    }


    fs << "PCA" << "{pca";//open struct (NODE_MAP)
    fs << "mean"<< myPCA.mean;
    fs << "eigenvectors" << myPCA.eigenvectors(Range(0,keepN+1), Range::all());
    fs << "eigenvalues" << myPCA.eigenvalues(Range(0,keepN+1), Range::all());
    fs << "}";

    std::cout << keepN+1 <<  " vectors written"<< std::endl;

    return 0;
}

double normalizar(const Mat&in, Mat &m)
{

    //smooth
    //    Gaussian

    //Mat k = m.clone();

    GaussianBlur(in, m, Size(3,3),1,1);

    Scalar me, st;

    meanStdDev(m, me, st);
    m -= me;

    if(st.val[0] > DEFAULT_MIN_STD_VALUE)
 	m = m / st.val[0];
    
// #ifdef DEBUG
//     else
// 	std::cerr << "Low std value!: "<< st.val[0] << std::endl;
// #endif		
    
    
    return st.val[0];

}



/* 
   1.- Read shape and apperance models 

   2.- Read samples (shape and texture)

   3.- Extract values of b_g and b_s for the samples.

   4.- For each b_{s_i}, modify it by \alpha, and
   
   4.1.- For each shape s_j, regenerate it with the modified b_{s_i}
   value from the original
   4.2.- Extract the corresponding texture from the image.
   4.3.- Project the texture on the PCA -> b'_g_j.
   4.4.- w_i = sqrt(sum(( b'_g_j - b_g_j)^2 )) 
   
   

*/



int main(int argc, char *argv[])
{


    cliparams_t params;
    
    if(argc < 2){
	std::cerr << "Not enough arguments! Use -h for help" << std::endl;    
	return -1;
    }

    int ret = parse_cli(argc, argv, params);

    if(params.inputs.size()<2 || ret){
	std::cerr << "Not enough arguments! Use -h for help" << std::endl;    
	return -1;
    }

    //***************
    //1.- Read shape and apperance models 

    AppearanceBase app;
    Shape shape;
    
    if(app.load(params.app_file))
	return -1;
    
    cout << "Apperance Base Loaded OK"<<endl;
    
    if(shape.m_base.load(params.shape_file))
	return -1;
    
    cout << "Shape Base Loaded OK"<<endl;
    

    //***************
    //  2.- Read samples (shape and texture) 

        
    vector<string> imgfiles, marks;
    int read = read_image_mark_files(params.inputs, params.use_bioid, imgfiles, marks);
    if(read <=0){
	std::cerr << "Error reading images and mark files"<<std::endl;
	return -1;
    }
    
    std::cout << imgfiles.size() << " images and marks read"<<std::endl;


    //***************
    // 3.- Extract values of b_g and b_s for the samples.
   
    vector<pair<Mat,Mat> > b_s; //list of parameters corresponding to the
		     //(unmodified) shapes
    
    vector<Mat> b_g; //list of parameters corresponding to the texture

    vector<Mat> images; //list of images in the set

    BioIDParser bIDparser;


    Mat accumm = Mat::zeros(1, shape.m_base.eigenvalues.rows,CV_32F);
    Mat ab = Mat::zeros(1, app.m_base.eigenvalues.rows,CV_32F);

    for(uint ai=0; ai< imgfiles.size(); ai++){

	Mat g, s, d;
	Mat img, points;
	vector<Mat> patches;

	if(params.use_bioid){
	    bIDparser.load(marks.at(ai).c_str());
	    points = (bIDparser.get_points());	    
	}
	else
	    points = (CvMat*)cvLoad(marks.at(ai).c_str());

	img = imread(imgfiles.at(ai), CV_LOAD_IMAGE_UNCHANGED);
	img = Mat_<float>(img);
	img = img * 1./255;
	
	
	int read = extract_patches(img, points, patches, 
				   app.m_sz);

	if((uint)read != patches.size() || ((int)patches.size() != app.m_patch_count)){//incomplete shape
	    continue;	    
	}

	Mat p = Mat::zeros (app.m_sz.height*app.m_patch_count, app.m_sz.width,
	       CV_32F);

	for(uint k=0; k<patches.size();k++){
	    Mat m = p(Range(k*app.m_sz.height, (k+1)*app.m_sz.height),
		      Range::all());

	    if(app.m_normalized)
		normalizar(patches.at(k), m);
	    else
		patches.at(k).copyTo(m);

	}


	p = p.reshape(0,1);
	g = app.m_base.project(p);
	//should I limit the values of the parameters to 3 std?? - I
	//guess so, because the shape parameters are cap.	
// 	g = min(g, 3*app.m_eig_sqrt);
// 	g = max(g, -3*app.m_eig_sqrt);
	
	shape.set(points, false);
	shape.get_parameters(s,d);
	b_s.push_back(make_pair(s,d));

	
	accumm = accumm + s;
	ab = ab + g;


	b_g.push_back(g);

	images.push_back(img);
       
    }
    
    cout << "Optimal parameters extracted: "<< images.size()<< " samples"<<endl;

    cout << "Accummm: " <<endl<<accumm<<endl;
    cout << "Acc_g: " <<endl<<ab<<endl;

    //***************
    //   4.- For each b_{s_i}, modify it by \alpha, and
    
    //    4.1.- For each shape s_j, regenerate it with the modified b_{s_i}
    //    value from the original
    //    4.2.- Extract the corresponding texture from the image.
    //    4.3.- Project the texture on the PCA -> b'_g_j.
    //    4.4.- w_i = sqrt(sum(( b'_g_j - b_g_j)^2 )) 
    
    int nparams = shape.m_base.eigenvalues.rows;

    Mat w = Mat::zeros(1, nparams, CV_32F);

    for(uint k=0; k < (uint)nparams; k++){
	uint r = 0;
	double wk = 0;
	double disp =  params.alpha * shape.m_base.m_eig_sqrt.at<float>(0,k); //displacement

	for(uint ai = 0; ai< images.size(); ai++){
	
	    vector<Mat> patches;

	    //modify the k-th parameter in the shape
	    
	    Mat sprime = b_s.at(ai).first.clone();
	    Mat d = b_s.at(ai).second;
	    Mat gprime;
	    
	    sprime.at<float>(0,k) = sprime.at<float>(0,k) + disp;
	    
	    shape.set_parameters(sprime, d);

	    //generate the shape

	    Mat pts = shape.get();

	    //extract texture

	    int read = extract_patches(images.at(ai), pts, patches, app.m_sz);

	    if(read!= (int)patches.size()){
		continue;
	    }

	    //project texture
	    Mat p (Size( app.m_sz.width, app.m_sz.height*app.m_patch_count),
		   CV_32F);
	    
	    for(uint t=0; t<patches.size();t++){
		Mat m = p(Range(t*app.m_sz.height, (t+1)*app.m_sz.height),
			  Range::all());
		
		if(app.m_normalized)
		    normalizar(patches.at(t), m);
		else	
		    patches.at(t).copyTo(m);
	    }
	    
	    p = p.reshape(0,1);
	    gprime = app.m_base.project(p);
	    //should I limit the values of the parameters to 3 std?? - I
	    //guess so, because the shape parameters are cap.	
// 	    gprime = min(gprime, 3*app.m_eig_sqrt);
// 	    gprime = max(gprime, -3*app.m_eig_sqrt);
	    

	    //get the difference between original parameters and new values.
	    
	    wk = wk + norm(gprime - b_g.at(ai));

	    r++;

	}

	wk = sqrt(wk/r);
	wk = wk / fabs(disp);

	w.at<float>(0,k) = wk;

	cout << "Weight for parameter "<<k<< ": "<< wk<<endl;
	
    }

    images.clear();


    //***************
    //5.- run PCA con the set of parameters
    
    Mat data(b_s.size(), shape.m_base.eigenvalues.rows + 
	     app.m_base.eigenvalues.rows, CV_32F);
    
    for(uint ai=0; ai<b_s.size(); ai++){

	Mat p = data(Range(ai,ai+1), Range(0, shape.m_base.eigenvalues.rows));
	b_s.at(ai).first.copyTo(p);
	//	p = p.mul(w);

	p = data(Range(ai,ai+1), Range(shape.m_base.eigenvalues.rows, data.cols));
	b_g.at(ai).copyTo(p);

    }
    
    {
	//	Mat tt = data.colRange(0,shape.m_base.eigenvalues.rows);
	Mat ttc;

	reduce(data, ttc, 0, CV_REDUCE_SUM);

	cout << ttc<<endl;

    }
    PCA myPCA (data, Mat(), CV_PCA_DATA_AS_ROW);

    //the mean in myPCA should be all zero. Because of rounding errors
    //it is not zero, so we set it.

    cout << myPCA.mean<<endl;

    myPCA.mean = Scalar(0);


    //***************
    //6.- Save the matrix and the coupled PCA

    FileStorage fs(params.ofile, FileStorage::WRITE);
    if(fs.isOpened()){
	cout << "Matrix and coupled base written to "<<params.ofile<<endl;
	
	fs << "CoupledBase" <<"{CoupledBase";
	fs<< "weight_matrix"<< w;

	savePCA(fs, myPCA, params.keep_pca, params.max_components);

	fs << "}";
    }
    else{
	cerr << "Error writing matrix to disk"<<endl;
	cout << w;
	return 1;
    }

    
   

    fs.release();
	

    return 0;
}
 
