// A software to estimate the value of the constant K in the cost
// function of CLM.
 /*
    CLM_uah - Constrained Local Model (CLM) sample code. A (partial)
    implementation of D. Christinacce and T. Cootes algorithm.

    Jesus Nuevo-Chiquero, University of Alcal. 2010
    jesus.nuevochiquero@gmail.com

    This file is part of 'CLM_uah'.

    CLM_uah is free software: you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    CLM_uah is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with CLM_uah.  If not, see <http://www.gnu.org/licenses/>.

*/
#include <vector>
#include <stdlib.h>
#include <algorithm>
#include <iostream>
#include <fstream>
#include <iterator>

#include <stdexcept>

#include <boost/program_options.hpp>

#include <gsl/gsl_multimin.h>
#include <cxcore.h>
#include <cv.h>
#include <highgui.h>

#include "basic_utils.hh"

#include "Shape.hh"
#include "AppearanceBase.hh"

#include "config.h"

const char *const svn_str = "$Id: estimate_K.cpp 946 2010-11-28 18:14:49Z jnuevo $";
const char *const svn_rev = "$Rev: 751 $";


using namespace cv;
using namespace std;
namespace po=boost::program_options;

#define DEFAULT_MIN_STD_VALUE 1e-2


typedef struct cliparams_t{
    //    string output;
    vector<string> inputs;
    string shape_file, app_file;   

}cliparams_t;


int parse_cli(int argc, char *argv[], cliparams_t &clarg)
{

    
  try{

	po::options_description generic("Generic options");
	generic.add_options()
	    ("version,v", "print version string")
	    ("help,h", "produce help message")    
	    ;

	po::options_description config("Configuration (Required)");
        config.add_options()
	    ("app_model,a",po::value<string>(&clarg.app_file), "Appearance model description file")
	    ("shape_model,s", po::value<string>(&clarg.shape_file), "Shape model (PDM) description file")
	    ;

	//	po::options_description output("Output");
// 	output.add_options()
// 	    ("output,o", po::value<std::string>(), "Output folder")
// 	    ;
	
	// Hidden options, will be allowed both on command line and
        // in config file, but will not be shown to the user.
        po::options_description hidden("Hidden options");
        hidden.add_options()
            ("input-files", po::value< std::vector<std::string> >(&clarg.inputs), "input files")
            ;
	
  	po::positional_options_description p;
        p.add("input-files", -1);

	po::options_description visible;
        visible.add(generic).add(config);
	

	po::options_description cmdline_options;
        cmdline_options.add(generic).add(config).add(hidden);

        po::variables_map vm;
        store(po::command_line_parser(argc, argv).
              options(cmdline_options).positional(p).run(), vm);
	
        notify(vm);
	
	if(vm.count("version")){
	    std::cout << "Version: "<<CLM_uah_VERSION_MAJOR<<"."
		      <<CLM_uah_VERSION_MINOR<<"."
		      <<CLM_uah_VERSION_PATCH<<std::endl;
	    std::cout << "SVN rev. : " <<svn_rev << std::endl;
	}

	if (vm.count("help")) {
	    std::cout << "Estimate the value of the constant K in the cost function to minimize with CLM (see fitting.cpp)"<<std::endl;
	    std::cout << "Usage: estimate_K options <FILES>"<<std::endl;
	    std::cout << visible << std::endl;
	    std::cout << "FILES are files with patch and points data generated with extract_patches"<<endl;
            return 1;
        }

	if(!vm.count("input-files")){
	    std::cout << "No input files"<<std::endl;
	    return 1;
	}

  }
  catch(std::exception& e)
      {
	  std::cout << e.what() << "\n";
	  return 1;
      }    
  
  return 0;

}


double normalizar(const Mat k, Mat &m)
{

    //smooth
    //    Gaussian

    //    Mat k = m.clone();

    GaussianBlur(k, m, Size(3,3),1,1);

    Scalar me, st;

    meanStdDev(m, me, st);
    m -= me;

    if(st.val[0] > DEFAULT_MIN_STD_VALUE)
 	m = m / st.val[0];
    
#ifdef DEBUG
    else
	std::cerr << "Low std value!: "<< st.val[0] << std::endl;
#endif			    
    
    return st.val[0];

}



double costR(const Mat& mat, const AppearanceBase& ab, Mat& acc)
{

    double cost = 0;

    //project the image over the PCA in AppBase

    Mat m = mat.reshape(0,1);

    Mat proj = ab.m_base.project(m);

    acc = acc+proj;

  //   cout << proj << endl;

    //Limit the value of the projection with the eigenvalues
    Mat sq; 
    sqrt(ab.m_base.eigenvalues, sq);
    //  proj = proj.reshape(0,proj.cols);
     proj = min(proj, 3*sq.t());
     proj = max(proj, -3*sq.t());
    

    //reproject

    Mat bp = ab.m_base.backProject(proj);

    // cost = 1 - NCC(reprojection,image)
    
    matchTemplate(m, bp, proj, TM_CCOEFF_NORMED);

    cost = 0.5 - 0.5*proj.at<float>(0,0);

    return cost;

}


double costS(const Mat& shape_, Shape& sb, Mat&acc)
{

    Scalar cost;
    Mat shape = shape_.clone();
    
    sb.set(shape);
    Mat proj, st;

    sb.get_parameters(proj, st);

    acc = acc + proj;

  //   Mat sq; 
//     sqrt(sb.eigenvalues, sq);
    //cost = sum(params^2 / eigvals)

    pow(proj, 2, proj);
    proj = proj.reshape(0,proj.cols);
    cost = sum(proj/sb.m_base.eigenvalues);

    return cost[0];

}


int main(int argc, char *argv[])
{

    cliparams_t params;
    vector<double> ks;
    
    if( argc < 2){
	std::cerr << "Not enough arguments! Use -h for help" << std::endl;    
	return -1;
    }
    
    if(parse_cli(argc, argv, params))
	return 1;
    

    //load the pdm and app model

    AppearanceBase app;
    Shape shape;

    if(app.load(params.app_file))
	return -1;

    cout << "Apperance Base Loaded OK"<<endl;

    if(shape.m_base.load(params.shape_file))
	return -1;
	
    cout << "Shape Base Loaded OK"<<endl;


    Mat ab = Mat::zeros(1, app.m_base.eigenvalues.rows, CV_32F);
    Mat as = Mat::zeros(1, shape.m_base.eigenvalues.rows, CV_32F);
    

    //foreach image and shape:
    for(uint ai=0; ai<params.inputs.size();ai++){
	
	Mat patches (Size( app.m_sz.width, app.m_sz.height*app.m_patch_count),
		     CV_32F);
	Mat points;

	//load the extracted patches and points
	FileStorage fs(params.inputs.at(ai), FileStorage::READ);
	if(!fs.isOpened()){
	    cerr<< "Error abriendo "<<params.inputs.at(ai)<<endl;
	    
	}
	else{
	    int npatches, nOK;
	    char str[128];
	    fs["total_patches"] >> npatches;
	    fs["patches_written"] >> nOK;
	    
	    if(nOK != npatches)
		continue;

	    try{
		read(fs["points"],points);
	    }
	    catch(Exception e){
		cerr << "Points not found "<< e.err <<endl;
		continue;
	    }

	    for(int k=0; k<npatches;k++){
		sprintf(str, "patch_%03d",k);
		Mat m, p = patches(Range(k*app.m_sz.height, (k+1)*app.m_sz.height),
				Range::all());
		read(fs[str],m);

		if(m.size()!=p.size()){
		    cerr<<"Patch and texture base sizes do not match"<<endl;
		    return -1;
		}		    
		
		normalizar(m, m);

		m.copyTo(p);		
	    }
	    
	//     namedWindow("estimateK",1);
// 	    imshow("estimateK",patches);
// 	    unsigned char key = static_cast<unsigned char> (waitKey(wait_time));
// 	    if(key == '0' || key == ' '){
// 		wait_time = 0;
// 	    }		
// 	    else if(key == '1'){
// 		wait_time = 2;
// 	    }
// 	    else if(key == '2'){
// 		wait_time = 10;
// 	    }
// 	    else if(key == '3'){
// 		wait_time = 100;
// 	    }
// 	    else if(key == 'd'){
// 		Mat d = patches(Rect(0,0, 10,10));
// 		cout << d <<endl;
// 	    }
	    
	    //eval costS, costR
	    
	    double cs = costS(points, shape, as);
	    double cr = costR(patches, app, ab);
	    
	    double alpha = cs/cr;
	    
	    ks.push_back(alpha);
	    
	    Mat pr, pr1;
	    shape.get_parameters(pr, pr1);

	    if(ai<10)
		cout << pr<<endl;

	    //alpha = costS/costR;
	    //	cout << "Alpha: "<<alpha<<endl;
	}
    }

    cout << ab<<endl;
    cout << as<<endl;

    double m=0, med;
    for(vector<double>::iterator it = ks.begin(); it != ks.end(); it++){
	m += *it;
    }

    cout << "Mean value: " << m/ks.size() << endl;

    sort(ks.begin(), ks.end());
    if(ks.size()%2)
	med = ks.at(1 + (ks.size()/2));
    else{
	med = ks.at(ks.size()/2 -1) + ks.at(ks.size()/2 +1);
	med /= 2;
    }

    cout << "Median: " << med << endl;

    return 0;


}
