 /*
    CLM_uah - Constrained Local Model (CLM) sample code. A (partial)
    implementation of D. Christinacce and T. Cootes algorithm.

    Jesus Nuevo-Chiquero, University of Alcal. 2010
    jesus.nuevochiquero@gmail.com

    This file is part of 'CLM_uah'.

    CLM_uah is free software: you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    CLM_uah is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with CLM_uah.  If not, see <http://www.gnu.org/licenses/>.

*/
#include <sys/time.h>
#include <iostream>
#include <vector>

#include "cxcore.h"
#include "cv.h"
#include <math.h>

#include <gsl/gsl_multimin.h>

#include "fitting.hh"


#define CLM_SIMPLEX_K 1./150 //computed with estimate_K
#define SIMPLEX_SCALE_STEP 5
#define SIMPLEX_ROTATION_STEP 0 // M_PI/90.
#define SIMPLEX_TRANSLATION_STEP 3

#define SIMPLEX_MAX_ITER 500
#define CLM_MAX_FIT_ERROR 0.07

#define SIMPLEX_EIGENVALUE_ALPHA 1.

#define CLM_PRECOMPT_RATIO 2.5 //size of the recomputed area, in terms of the patch size

using namespace cv;
using namespace std;

//point evaluated out of the precalculated area.
typedef struct eval_pt_t{
    Point p;
    double val;
    unsigned int count;
}eval_pt_t;

//A structure to contain the parameters and data of the minimization
typedef struct simplexData_t{
    Mat img;
    Mat tmplt;//concatenated templates
    Size sz; //patch size
    vector<Mat> match_pre;//precalculated surface responses around the
			  //initial positions
    vector<Rect> match_rect;//top left corner of the precalculated
			   //surface responses
    Shape shape;
    
    vector<vector<eval_pt_t> >evpts;

    int total_eval;

#ifdef DEBUG
    bool print;
#endif
    
}simplexData_t;


double eval_zncc(const Mat& img, const Mat& tmp)
{

    Scalar m1, m2, d;
  
    m1 = mean(img);
    m2 = mean(tmp);
    
    Mat a1 = img.clone();
    a1 -= m1;
    Mat a2 = tmp.clone();
    a2 -= m2;
    
    d = a1.dot(a2);

    double n1 = norm(a1), n2 = norm(a2);
    
    return 0.5 -(d.val[0] / (n1*n2))/2;
    
}

/** 
 * Evaluate the objective function. 
 *
 * This function uses the CLM_SIMPLEX_K constant, calculated with
 * "estimate_K". It is called by the gsl minimizer.
 * 
 * @param v The vector containing the variables of the minimization
 * @param params The parameters (a simplexData_t struct)
 * 
 * @return The function value.
 */
double eval_obj_function(const gsl_vector *v, void *params)
{

    simplexData_t *data = (simplexData_t*) params;

    double val = 0;
    int counter = 0;

    data->total_eval ++;

    //instantiate shape from parameters

    Mat_<float> p = Mat_<float>( 1, v->size);
    Mat_<float> s  = p.colRange(0, v->size-4);
    Mat_<float> st = p.colRange( v->size-4, p.cols);
    
    for(uint ai=0; ai<v->size; ai++)
	p(0,ai) = gsl_vector_get(v, ai);

    data->shape.set_parameters(s,st);

    Mat_<float> pts = data->shape.get();
 
    Rect iR (Point(data->sz.width/2,data->sz.height/2),
	     data->img.size());
    iR -= data->sz;

   
    for(int ai=0; ai < pts.cols; ai++){
	Point p = Point(pts(0,ai), pts(1,ai));
	//check points are in image

	if(!iR.contains(p)){
	    val = val + 1; ////ignore point
	}
 
	//search for the points in the precalculated areas
	else if(data->match_rect.at(ai).contains( p )){

	    p = p - data->match_rect.at(ai).tl();
	    val = val + data->match_pre.at(ai).at<float>(p);
	    
	    counter++;
	}
		
	//calculate if outside the area.
	else{
	    //ZNCC
	    bool found = false;
	    for(uint k=0; k < data->evpts.at(ai).size(); k++){
		eval_pt_t& e = data->evpts.at(ai).at(k);
		if(e.p == p){
		    if(!isnan(e.val)){
			val += e.val;
		    }
		    e.count++;
		    found = true;
		    break;
		}
	    }

	    if(found)
		continue;

	    Rect r (Point(p.x - data->sz.width/2, p.y - data->sz.height/2), data->sz);

	    double k = eval_zncc(data->img(r), data->tmplt.rowRange(ai*data->sz.height, (ai+1)*data->sz.height));
	    
	    if(!isnan(k)){
		val += k;
		counter++;
	    }

	    eval_pt_t ep;
	    ep.p = p;
	    ep.val = k;
	    ep.count = 1;
		
	    data->evpts.at(ai).push_back(ep);
	    
	}
    }

    if(!counter)
	val = pts.cols;//1 for all points

    

    //error from shape parameter values

    pow(s, 2, s);
    s = s.reshape(s.cols);
    Scalar cost = sum(s/data->shape.m_base.eigenvalues);

    cost.val[1] = val + CLM_SIMPLEX_K * cost.val[0];

#ifdef DEBUG
    if(data->print){
	cout << "val(ncc): " <<val << ", shape: "<< CLM_SIMPLEX_K *cost.val[0]<<endl;
	data->print = false;
    }
#endif
    
    return cost.val[1];

}



unsigned int simplex_fitting(simplexData_t& data, double &fit)
{

    const gsl_multimin_fminimizer_type *T = gsl_multimin_fminimizer_nmsimplex;
    gsl_multimin_fminimizer *s = NULL;
    gsl_vector *ss, *x;
    //    gsl_vector_view view;
    gsl_multimin_function minex_func;
    int status;

    Mat bg, st;

    data.shape.get_parameters(bg, st);

    unsigned int np = bg.cols + st.cols;
    unsigned int iter = 0;
    double fval=1e6, size, init_size;
    double max_error = CLM_MAX_FIT_ERROR * data.shape.m_base.eigenvectors.cols/2;

    x = gsl_vector_alloc(np);
    ss = gsl_vector_alloc (np);
   
    for(int ai=0; ai< bg.cols; ai++){
	gsl_vector_set(x, ai, bg.at<float> (0,ai));
	gsl_vector_set(ss, ai, data.shape.m_base.m_eig_sqrt.at<float>(0,ai) * SIMPLEX_EIGENVALUE_ALPHA);
    }
   
    for(int ai=0; ai< 4; ai++){
	gsl_vector_set(x, ai+ bg.cols, st.at<float>(0,ai));	
    }

    //scale step
    gsl_vector_set(ss, np-4, SIMPLEX_SCALE_STEP);
    gsl_vector_set(ss, np-3, SIMPLEX_ROTATION_STEP);
    gsl_vector_set(ss, np-2, SIMPLEX_TRANSLATION_STEP);
    gsl_vector_set(ss, np-1, SIMPLEX_TRANSLATION_STEP);
    
    minex_func.f = &eval_obj_function;
    minex_func.n = np;
    minex_func.params = &data;

    s = gsl_multimin_fminimizer_alloc (T, np);
    gsl_multimin_fminimizer_set (s, &minex_func, x, ss);


    s = gsl_multimin_fminimizer_alloc (T, np);
    gsl_multimin_fminimizer_set (s, &minex_func, x, ss);
    

    init_size = gsl_multimin_fminimizer_size (s);

    do{

	status = gsl_multimin_fminimizer_iterate(s);
	    
	if (status)
	    break;
	size = gsl_multimin_fminimizer_size (s);

	//run at least a few iterations
	status = gsl_multimin_test_size(size, init_size/20.);
	
	if(status != GSL_SUCCESS && s->fval < max_error)	    
	    status = GSL_SUCCESS;
	else
	    status = gsl_multimin_test_size(size, init_size/30.);

#ifdef DEBUG
	//	status = gsl_multimin_test_size (size, 1e-5);
	if( (fval - s->fval)>0.1*fval){
	    cout << "iteration: "<<iter <<", size: "<<size<< ", fit: "<< s->fval << endl;
	    data.print = true;
	    fval = s->fval;
	}
#endif
	iter++;

	//size check is not appropiate, because of the different step
	//sizes	
	


    }while( iter < SIMPLEX_MAX_ITER/3 || (status == GSL_CONTINUE && iter < SIMPLEX_MAX_ITER));

    cout << "iteration: "<<iter <<", size: "<<size<< ", fit: "<< s->fval << endl;

    for(int ai=0; ai< bg.cols; ai++){
	bg.at<float>(0,ai) = gsl_vector_get(s->x, ai);
    }

    for(int ai=0; ai< st.cols; ai++){
	st.at<float>(0,ai) = gsl_vector_get(s->x, ai + bg.cols);
    }
    
    data.shape.set_parameters(bg,st);

    fit = s->fval;


#ifdef DEBUG
    uint k = 0;
    for(uint ai=0; ai<data.evpts.size(); ai++){
	for(uint b= 0; b<data.evpts.at(ai).size();b++){

	    k += data.evpts.at(ai).at(b).count;

	}
    }

    cout << "Total out precomputed: "<< k <<endl;
#endif


    return iter;

}




double fit_CLM(Mat& img, Mat& tmplt, Shape& pdm)
{

    Mat points = pdm.get();

    simplexData_t s;

    s.img = img;
    s.tmplt = tmplt;
    s.sz = Size(tmplt.cols, tmplt.rows/points.cols);
    
    s.evpts.resize(points.cols);

    s.total_eval = 0;

    //get the shape parameters

    s.shape = pdm;
   
    //pre-calculate the correlation in an area around each point. The
    //size of the area is 0.5 *width of the patch on each side
    
    Rect iR (Point(0,0), img.size());

    for(int ai=0; ai<points.cols; ai++){

	Mat pre = Mat((CLM_PRECOMPT_RATIO-1)*s.sz.height +1, (CLM_PRECOMPT_RATIO-1)*s.sz.width +1, img.type());
	Point tl;

	Rect r (points.at<float>(0,ai) - CLM_PRECOMPT_RATIO*s.sz.width/2,
		points.at<float>(1,ai) - CLM_PRECOMPT_RATIO*s.sz.height/2,
		CLM_PRECOMPT_RATIO*s.sz.width, CLM_PRECOMPT_RATIO*s.sz.height);
	
	Rect maxR = iR & r; 

	if(maxR.width <= 0 || maxR.height <= 0){
	    //completely out of image. ignore the point.
	    pre = 1;	    
	}
	else if(!(r == maxR)){ //part of the area out of image - TODO
	//     matchTemplate(img(maxR), tmplt(Range(ai*sz.height,(ai+1)*sz.height), Range::all),
// 			  pre(Rect(0,0,maxR.width, maxR.height)), TM_CCOEFF_NORMED);
	    
// 	    pre(Rect(maxR.width, maxR.height, pre.width, pre.height)) = 1;

	    pre = 1;
	}
	else{
	    matchTemplate(img(maxR), tmplt.rowRange(ai*s.sz.height, (ai+1)*s.sz.height),
			  pre, TM_CCOEFF_NORMED);
	}
	
	pre = -.5* pre;
	pre += Scalar(0.5);
	//	pre = 0.5*pre;
	
	s.match_pre.push_back(pre);
	
	r = Rect(points.at<float>(0,ai) - (CLM_PRECOMPT_RATIO-1)*s.sz.width/2,
		 points.at<float>(1,ai) - (CLM_PRECOMPT_RATIO-1)*s.sz.height/2,
		 (CLM_PRECOMPT_RATIO-1)*s.sz.width, (CLM_PRECOMPT_RATIO-1)*s.sz.height);

	s.match_rect.push_back(r);

    }

    //localize best point modifying (tx,ty)
    Mat acc = s.match_pre.at(0).clone();
    for(uint ai=1; ai< s.match_pre.size();ai++){
	acc = acc + s.match_pre.at(ai);
    }

    Point m0;
    minMaxLoc(acc, NULL,NULL, &m0);
    m0 = m0 - Point(acc.cols/2, acc.rows/2);    
    if(m0.x || m0.y){
	Mat p,d;

	s.shape.get_parameters(p,d);

	d.at<float>(0,2) += m0.x;
	d.at<float>(0,3) += m0.y;

	s.shape.set_parameters(p,d);
    }

    cout<<"punto ptimo: " << m0.x << " "<< m0.y<<endl;

    

    //run simplex
    double fit;
    uint iter = simplex_fitting(s, fit); 

    cout << "Iterations: " << iter<< ", total evaluations: "<< s.total_eval<< endl;

    points = s.shape.get();


    return fit;
      
}

