 /*    
    CLM_uah - Constrained Local Model (CLM) sample code. A (partial)
    implementation of D. Christinacce and T. Cootes algorithm.

    Jesus Nuevo-Chiquero, University of Alcal. 2010
    jesus.nuevochiquero@gmail.com

    This file is part of CLM_uah.

    CLM_uah is free software: you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    CLM_uah is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with CLM_uah.  If not, see <http://www.gnu.org/licenses/>.

*/
#include <time.h>

#include <sys/stat.h>
#include <sys/types.h>
#include <errno.h>

#include <vector>
#include <algorithm>
#include <iostream>
#include <iterator>
#include <fstream>

#include <stddef.h>
#include <stdio.h>
#include <sys/types.h>
#include <dirent.h>
#include <unistd.h>
#include <sys/wait.h>

#include <stdexcept>

#include <boost/program_options.hpp>
#include <gsl/gsl_multimin.h>

#include <cxcore.h>
#include <highgui.h>

#include "config.h"

#ifdef HAS_RAW_VIDEO_READER_H
#include "raw_video_reader.hh"
#endif


#include "basic_utils.hh"
#include "utils_track_loss.hh"
#include "drawing.hh"
#include "bioid_parser.hh"

#include "Shape.hh"
#include "AppearanceBase.hh"
#include "CoupledBase.hh"
#include "fitting.hh"


using namespace std;
using namespace cv;
namespace po=boost::program_options;

const char *const svn_str = "$Id: run_CLM.cpp 945 2010-11-28 18:03:55Z jnuevo $";
const char *const svn_rev = "$Rev: 945 $";


#define DEFAULT_MIN_STD_VALUE 1e-2

#define TEXTURE_SATURATION_K 1
#define COUPLING_SATURATION_K 3

typedef struct cliparams_t{

    string shape_file, app_file, app_pp_file;
    string cw_file;
    string input;

    string reference;

    bool save_results;
    string output_root;    

    unsigned int wait;
 
    vector<double> noise;

}cliparams_t;


int parse_cli(int argc, char *argv[], cliparams_t &clarg)
{

    
  try{

	po::options_description generic("Generic options");
	generic.add_options()
	    ("version,v", "print version string")
	    ("help,h", "produce help message")    
	    ;

	po::options_description config("Configuration (Required)");
        config.add_options()
	    ("app_model,a",po::value<string>(&clarg.app_file), "Texture model description file")
	    ("shape_model,s", po::value<string>(&clarg.shape_file), "Shape model (PDM) description file")
	    ;


	po::options_description config2("Configuration");
        config2.add_options()
	    ("coupled_weight,c", po::value<string>(&clarg.cw_file),"Shape parameter weights for model parameter coupling")
	    ("app_model_i,i", po::value<string>(&clarg.app_pp_file), "Per-patch texture model description file")
	    ("reference,r", po::value<string>(&clarg.reference), "File containing the ground-truth position of the points")	   
	    ("wait,w",po::value<unsigned int>(&clarg.wait)->default_value(0),"Wait time for waitKey()")
	    ("noise,n",po::value<vector<double> >()->multitoken(), "mean and standard deviation of the noise added to the initial "
"points.\n" "If only one value is given, it is taken as the stv, and the mean is set to 0")
	    ;

	po::options_description output("Output");
	output.add_options()
	    ("output,o",  "Generate output")
	    ;
	

	// Hidden options, will be allowed both on command line and
        // in config file, but will not be shown to the user.
        po::options_description hidden("Hidden options");
        hidden.add_options()
            ("input-files", po::value< string >(&clarg.input), "input file (image or movie)")
            ;
	
  	po::positional_options_description p;
        p.add("input-files", 1);

	po::options_description visible;
        visible.add(generic).add(config).add(config2).add(output);
	

	po::options_description cmdline_options;
        cmdline_options.add(generic).add(config).add(config2).add(output).add(hidden);

        po::variables_map vm;
        store(po::command_line_parser(argc, argv).
              options(cmdline_options).positional(p).run(), vm);

	notify(vm);


	if(vm.count("version")){
	    std::cout << "Version: "<<CLM_uah_VERSION_MAJOR<<"."
		      <<CLM_uah_VERSION_MINOR<<"."
		      <<CLM_uah_VERSION_PATCH<<std::endl;
	    std::cout << "SVN rev. : " <<svn_rev << std::endl;
	}

	if (vm.count("help")) {
	    std::cout << "Usage: run_CLM <options> FILE"<<std::endl;
	    std::cout << visible << std::endl;
	    std::cout << "FILE is a video or an image"<<endl;
            return 1;
        }

	if(!vm.count("app_model")){
	    cerr << "ERROR: An appearance model file is needed"<<endl;
	    return -1;
	}

	if(!vm.count("shape_model")){
	    cerr << "ERROR: A shape model file is needed"<<endl;
	    return -1;
	}

	if( vm.count("noise") )	{
	    clarg.noise = vm["noise"].as< vector<double> >();
	    int n = clarg.noise.size();
	    if(n==1){//I suppose it is the std, and the mean is 0
		clarg.noise.push_back(clarg.noise.at(0));
		clarg.noise.at(0) = 0;
	    }

	    cout << "Added noise (mean, stv): " << clarg.noise.at(0) << ", " << clarg.noise.at(1)<<endl;

	}

	if(vm.count("output")){
	    for(uint ai=0; ai<1e5;ai++){
		char file[128];
		sprintf(file, "clm_results_data_%05d.avi", ai);

		if(access (file, F_OK)){
		    sprintf(file,"clm_results_data_%05d",ai);
		    clarg.output_root = file;
		    clarg.save_results = true;
		    break;
		}
	    }
	    if(!clarg.output_root.length()){
		std::cerr << "Output file could not be created, too many files in current folder"<<std::endl;
		exit(1);
	    }	
	}
       

	if(!vm.count("input-files")){
	    cerr<<"No input file"<<endl;
	    return -1;
	}

  }
  catch(std::exception& e)
      {
	  std::cout << e.what() << "\n";
	  return 1;
      }    
  
  return 0;

}


//normalize an image. The algorithm works better with some blurring.
double normalizar(Mat &m)
{

    Mat k = m.clone();

    GaussianBlur(k, m, Size(3,3),2,2);

    Scalar me, st;
 
    meanStdDev(m, me, st);
    m -= me;

    if(st.val[0] > DEFAULT_MIN_STD_VALUE)
 	m = m / st.val[0];
 
#ifdef DEBUG
    else{
	std::cerr << "Low std value!: "<< st.val[0]<<", mean"<<me.val[0] << std::endl;
    }
#endif			    
       
 
    return st.val[0];
}



/** 
 * Tentatively place the shape inside the given rectangle. The shape
 * is centered, and the inter-eye distance is set to half the width of
 * the rectangle.
 *
 * This function is model specific: if you use models with a different
 * point order or position, check if it fits your model and change it
 * accordingly
 * 
 * @param shape A shape. It is modified by the function.
 * @param r The rectangle
 */
void place_inside_rectangle(Shape shape, Rect r)
{

    Mat pts = shape.get();
    Mat b, s;

    shape.get_parameters(b,s);

    //make inter-eye distance be half the width of the square

    double d = pts.at<float>(0,1) - pts.at<float>(0,0);

    double k = (0.4*r.width);

    s.at<float>(0,0)  *= k / (d);
    //center the model to the center of the square

    Point c = (r.tl()+r.br());

    s.at<float>(0,2) = c.x/2;
    s.at<float>(0,3) = c.y/2;


    //limit the value of the parameters so it becomes more neutral:

    b = min(b, shape.m_base.m_eig_sqrt);
    b = max(b, -shape.m_base.m_eig_sqrt);

    shape.set_parameters(b,s);
    

}


/** 
 * Extract patches from \a _img around the points given by \a points
 * 
 * @param _img The image to extract from
 * @param points The points where the patches are
 * @param sample An image with the patches concatenated
 * 
 * @return The number of patches extracted.
 */
int extract_sample(const Mat& _img, const Mat& points, Mat& sample)
{

    Size sz = Size(sample.cols, sample.rows/points.cols);
    Rect iR (Point(0,0), _img.size());
    
    int out = 0;

    Mat img;

    GaussianBlur(_img, img, Size(3,3),1,1);

    for(int ai=0; ai<points.cols; ai++){

	Rect r (points.at<float>(0,ai) - sz.width/2,
		points.at<float>(1,ai) - sz.height/2,
		sz.width, sz.height);
	
	Rect maxR = iR | r;
	Mat s = sample(Range(ai*sz.height, (ai+1)*sz.height), Range::all());

	if(maxR == iR){//patch is in the image
	    img(r).copyTo(s);
	    normalizar(s);
	}
	else{
	    s = Scalar(0);
	    out++;
	}
    }

    return points.cols - out;

}


int main(int argc, char *argv[])
{


    cliparams_t params;

    if( argc < 2){
	std::cerr << "Not enough arguments! Use -h for help" << std::endl;    
	return -1;
    }
    
    if(parse_cli(argc, argv, params))
	return 1;
    
    unsigned int wait_time = params.wait;

    
    CascadeClassifier haar_front("/usr/local/share/opencv/haarcascades/haarcascade_frontalface_default.xml");

    AppearanceBase app, app_pp;
    Shape shape;
    CoupledBase coupled;
    

    //0.- load the models
    try{
	if(app.load(params.app_file)){
	    cerr<<"Error reading file, or file missing: "<<params.app_file<<endl;
	    return -1;
	}
    }
    catch(...){
	cerr<<"Error reading file, or file missing: "<<params.app_file<<endl;
	return -1;
    }
    cout << "Apperance Base Loaded OK"<<endl;

    if(app.m_per_patch){
	cerr<< "AppBase is per-patch. This is not supported yet."<<endl;
	return -1;
    }
    
    
    try{
	if(shape.load( params.shape_file )){
	    cerr<<"Error reading file, or file missing: "<<params.shape_file<<endl;
	    return -1;
	}
    }
    catch(...){
	cerr<<"Error reading file, or file missing: "<<params.app_file<<endl;
	return -1;
    }
    cout << "Shape Base Loaded OK"<<endl;

    
    if(params.cw_file.length()){
	try{
	    if(coupled.load(params.cw_file)){
		cerr<<"Error reading file, or file missing: "<<params.cw_file<<endl;
		return -1;
	    }
	}
	catch(...){
	    cerr<<"Error reading file, or file missing: "<<params.cw_file<<endl;
	    return -1;
	}

	cout << "Coupled Base Loaded OK"<<endl;

	if(coupled.eigenvectors.cols != (shape.m_base.eigenvectors.rows + app.m_base.eigenvectors.rows)){
	    cerr<<"Bad Coupled Base: size does not match" << endl;
	    return -1;
	}
    }


    if(params.app_pp_file.length()){
	try{
	    if(app_pp.load(params.app_pp_file)){
		cerr<<"Error reading file, or file missing: "<<params.app_pp_file<<endl;
		return -1;
	    }
	}
	catch(...){
	    cerr<<"Error reading file, or file missing: "<<params.app_pp_file<<endl;
	    return -1;
	}
	cout << "Per-patch Appearance Base Loaded OK"<<endl;

	if(!app_pp.m_per_patch){
	    cerr<< "Error: AppBasePP is not per-patch."<<endl;
	    return -1;
	}

    }

    VideoWriter writer;
    VideoCapture video;
    Mat lone_image;

   //1.- try to open the video

#ifdef HAS_RAW_VIDEO_READER_H
    RawVideoReader rawvideo;
    if(rawvideo.open( params.input.c_str() ))
#endif
	{
	    
	    video.open( params.input );
	    if(!video.isOpened()){
		lone_image = imread(params.input, 0);
		if(lone_image.empty()){
		    cerr <<params.input <<": No video or image file: exiting" << endl;
		    return -1;
		}
		cout<<"Image file read"<<endl;
	    }
	}
    
    /*********************************
     *
     *   2.- INITIALIZATION
     *
     *********************************/
    
    //2.- initialization
    //Options: a) read from file 
    //         b) use STASM (better accuracy than CLM, but much slower (=> plus some noise?) 
    //         c) use V&J, and place the points tentatively.
    
    Mat pts_t0;

    //get the image or the first frame
    if(lone_image.empty()){
	if(video.isOpened()){
	    video >> lone_image;
	    //video.set(CV_CAP_PROP_POS_FRAMES, 0);
	}
#ifdef HAS_RAW_VIDEO_READER_H
	else{	    
	    lone_image = cvarrToMat(rawvideo.query_frame(), true);
	    //  rawvideo.set_property(CV_CAP_PROP_POS_FRAMES, 0);
	}
#endif
    }

    {
	Mat img_sc;
	double sc = 1./max(lone_image.cols/320., lone_image.rows/240.);
	resize(lone_image, img_sc, Size(0,0), sc, sc);
	lone_image = img_sc;
    }

    if(params.reference.length()){
	//check if the file has BioID format (from the file extension)
	size_t sz = params.reference.find_last_of(".");
	if(params.reference.substr(sz+1)=="pts"){
	    BioIDParser bioid;
	    bioid.load(params.reference.c_str());
	    pts_t0 = bioid.get_points();
	    
	}
	else{
	    FileStorage fs(params.reference, FileStorage::READ);
	    
	    if(fs.isOpened()){
		read(fs.getFirstTopLevelNode(), pts_t0);
	    }
	}
	
    }
    else if(find_STASM()){//STASM
	std::cout << "Initializing with STASM"<<std::endl;
	
	CvMat *m = stasm_find_points(lone_image);
	if(CV_IS_MAT(m))
	    pts_t0 = m;
	else{
	    std::cerr << "STASM did not return a valid matrix"<<std::endl;
	    return -1;
	}
	//add noise.
	if(params.noise.size()==2){
	    Mat noise (pts_t0.size(), pts_t0.type());
	    randn(noise, Scalar(params.noise.at(0)), Scalar(params.noise.at(1)));
	    pts_t0 += noise;

	    //	    cout << norm(noise) << endl;
	    
	}
    } 
    else{//V&J
	vector<Rect> r;
	haar_front.detectMultiScale( lone_image, r, 1.1,3,0,Size(120,120));
	//TODO
	if(r.size()){
	    place_inside_rectangle(shape, r.at(0));
	    pts_t0 = shape.get();
	}
	
    }

    if(pts_t0.empty()){
	cerr << "No init/reference file,  STASM not found: exiting" << endl;
	return -1;
    }

    //2.1.- check the size of the shape base and the points
    
    if(shape.m_base.mean.size() != pts_t0.reshape(0,1).size()){
	cerr << "Sizes of shape model mean and initialization are not equal"<<endl;
	return -1;
    }


    if(params.save_results){
	//only video for now

	writer.open(params.output_root + ".avi", CV_FOURCC('M','P','4','2'), 25, lone_image.size());

    }

    /*********************************
     *
     *   3.- FITTING
     *
     *********************************/
    

    /*
      0.- Convert image (convertir imagen)
      
      1.- Extract patches from the images, given the current position
      of the shape (extraccin de parches de la imagen, segn posicin
      actual)

      2.- Obtain b_g projecting the patches over the texture vector
      base. (obtencin de b_g proyectando los parches sobre la base de
      textura)

      3.- Obtain 'c' from the texture coefficients and shape b_s
      (obtencin de c a partir de los parmetros de textura y forma
      b_s)

      4.- Recompute b_g (recalcular b_g y b_s?.)

      5.- Minimize the cost function f(b_s,st) = \sum_{i=1}^{n}
      R_i(\mathbf{X}_i) + K*sum_j^n \frac{b_{s_j}^2}{\lambda_j} using
      Nelder-Mead simplex algorithm.  (minimizar la funcin de coste mediante
      Nelder-Mead)

      6.- Repeat until convergence. (repetir hasta que converja.)

      Extras:
      a) Initialization. 
      b) Track losses detection (Deteccin de prdidas.)


      Improvements (TODO):
      a) Evaluate divergences using a local texture base (per-patch description of the texture).

    */

    namedWindow("CLM",0);
    cvResizeWindow("CLM",640,480);

    shape.set(pts_t0);
    //    pts_t0 = shape.get();

    Mat sample_ext = Mat(app.m_sz.height*app.m_patch_count, app.m_sz.width, 
			 app.m_base.mean.type());//extracted sample, from the image
    Mat img3, img_sc, img = lone_image;
    Mat pts;

    Mat b_s, b_g, b, c;//shape and appearance parameters
    unsigned int frame = 0;
    bool reset_stasm = false;
    
    vector<Rect> rect_vj;
    double time0, ttime;

    while(1){

	cout << "***********************"<<endl 
	     <<"Frame: "<< frame<<endl;
	
	//check image channels
	if(img.channels() == 3){
	    img3 = img;
	    cvtColor(img3, img, CV_BGR2GRAY);
	}
	else{
	    cvtColor(img, img3, CV_GRAY2BGR);
	}

	//convert to 32F
	img = Mat_<float>(img);
	img = img / 255.;	

	time0 = (double) getTickCount();

	pts = shape.get();
	
	// Step 1
	extract_sample(img, pts, sample_ext);

	// Step 2
	
	Mat m = sample_ext.reshape(0,1);
	b_g = app.m_base.project(m);       
	
	//2.1.- put limits on the b_g coefficients
 	b_g = min(b_g, TEXTURE_SATURATION_K*app.m_eig_sqrt);
 	b_g = max(b_g, -TEXTURE_SATURATION_K*app.m_eig_sqrt);
	
	{
	    Mat sat = (b_g == TEXTURE_SATURATION_K*app.m_eig_sqrt) | (b_g == - TEXTURE_SATURATION_K*app.m_eig_sqrt);
	    cout << "saturated parameters (texture):" << countNonZero(sat)<<endl;
	}
	
	//Step 3 .- (If a coupled base is given)

	if(!coupled.mean.empty()){
	    Mat aux;
	    shape.get_parameters(b_s, aux);

	    b.create(1, b_s.cols + b_g.cols, b_s.type());//only done the 1st time
	    
	    aux = b.colRange(0, b_s.cols);
	    aux = coupled.m_weight.mul(b_s);// copyTo(aux);
	    
	    aux = b.colRange(b_s.cols, b.cols);
	    b_g.copyTo(aux);
	    //aux = Scalar(0);
	    
	    
	    //Step 4
	    
	    c = coupled.eigenvectors * b.t();
	    c = min(c, COUPLING_SATURATION_K*coupled.m_eig_sqrt);
	    c = max(c, -COUPLING_SATURATION_K*coupled.m_eig_sqrt);
	    {
		Mat sat = (c == COUPLING_SATURATION_K*coupled.m_eig_sqrt) | 
		    (c == -COUPLING_SATURATION_K*coupled.m_eig_sqrt);
		cout << "saturated parameters (coupled):" << countNonZero(sat)<<endl;
	    }

	    b = coupled.m_eigvectors_t * c;

	    b = b.t();
	    b.colRange(b_s.cols, b.cols).copyTo(b_g);

	    b.colRange(0,b_s.cols).copyTo(b_s);
	    
	    //do not set the shape for now.
	    cout << "correccin de forma: "<< norm(b_s, b.colRange(0, b_s.cols))<<endl;

	}
	
	

	//Step 5.- minimize the cost function

	//5.1.- generate los parches
	
	app.m_base.backProject(b_g, sample_ext);

	sample_ext = sample_ext.reshape(0, app.m_sz.height*app.m_patch_count );

	fit_CLM(img, sample_ext, shape);
       

	pts = shape.get();

	ttime = (double)getTickCount() - time0;

	//draw + show
	if(rect_vj.size())
	    rectangle(img3, rect_vj.at(0).tl(), rect_vj.at(0).br(), CV_RGB(0,255,0));

	draw_points(img3, pts, CV_RGB(255,0,0));	
	draw_points(img3, pts_t0, CV_RGB(0,0,255));
	//draw rectangles
	draw_rectangles(img3, pts_t0, Size(1,1)+app.m_sz, CV_RGB(50,50,255));

	{
	    Mat c0, c1;
	    shape.get_parameters(c0, c1);

	    circle(img3, Point(c1.at<float>(0,2), c1.at<float>(0,3)), 2, CV_RGB(0,255,0));

	}

	//save for the next frame 
	pts.copyTo(pts_t0);

	imshow("CLM", img3);

	{
	    unsigned char key = static_cast<unsigned char> (waitKey(wait_time));
	    if(key == '0'){
		wait_time = 0;
// 		break;
	    }		
	    else if(key == '1'){
		wait_time = 2;
// 		break;
	    }
	    else if(key == '2'){
		wait_time = 10;
// 		break;
	    }
	    else if(key == '3'){
		wait_time = 100;
// 		break;
	    }
	    else if(key == ' '){
		wait_time = params.wait;
// 		if(smat_mask){
// 		    smat_mask= NULL;
		    
// 		}
// 		break;
	    }
	    else if(key == 's'){
		reset_stasm = true;

	    }
	    else if(key == 'q'){
		//		quit = true;
				
		break;
	    }
	}
	
	//save frame
	if(writer.isOpened())
	    writer << img3;

	//get new frame
	if(video.isOpened()){
	    video >> img;
	}
#ifdef HAS_RAW_VIDEO_READER_H
	else if(rawvideo.good()){	    
	    img = cvarrToMat(rawvideo.query_frame(), true);
	}
#endif
	else{
	    break;
	}

	if(img.empty()){
	    break;
	}
	
	rect_vj.clear();

	//scale new image down to 320x240
	double sc = 1./max(img.cols/320., img.rows/240.);
	resize(img, img_sc, Size(img.cols*sc, img.rows*sc));

	img = img_sc;
	
	//Initialize again if tracking is lost (detection is currently manual)
	if(reset_stasm && find_STASM()){

	    pts = stasm_find_points(img);
	    
	    if(!pts.empty()){
		shape.set(pts);
		
		cout << "Reset with STASM"<<endl;
	    }
	    reset_stasm = false;
	}

	else{
	    time0 = getTickCount();
	    haar_front.detectMultiScale( img, rect_vj, 1.1,3,0,Size(120,120));
	    if(rect_vj.size()){
		int out = 0;
		for(int ai=0; ai<pts.cols; ai++){
		    Point p (pts.at<float>(0,ai), pts.at<float>(1,ai));
		    if(!rect_vj.at(0).contains(p))
			out++;
		}

		if(out>pts.cols/3){//reposition
		    place_inside_rectangle(shape, rect_vj.at(0));


		    draw_points(img3, shape.get(), CV_RGB(255,255,0));	
		    imshow("CLM", img3);
		    waitKey(2*wait_time + 10);

		    
		}
	    }
	    ttime += (double)getTickCount() -  time0;
	}


	cout << "FPS: "<< getTickFrequency() / ttime <<endl;
	frame++;
    }
	    

    return 0;

}
